@@ -5,9 +5,9 @@
 ;; Author: Daniel Mendler <mail@daniel-mendler.de>
 ;; Maintainer: Daniel Mendler <mail@daniel-mendler.de>
 ;; Created: 2022
-;; Version: 1.4
-;; Package-Requires: ((emacs "27.1") (compat "30"))
-;; Homepage: https://github.com/minad/osm
+;; Version: 1.5
+;; Package-Requires: ((emacs "28.1") (compat "30"))
+;; URL: https://github.com/minad/osm
 ;; Keywords: network, multimedia, hypermedia, mouse
 
 ;; This file is part of GNU Emacs.
@@ -37,7 +37,7 @@
 ;; distances, search for locations by name and to open and display GPX
 ;; tracks.
 
-;; osm.el requires Emacs 27 and depends on the external `curl' program.
+;; osm.el requires Emacs 28 and depends on the external `curl' program.
 ;; Emacs must be built with libxml, libjansson, librsvg, libjpeg and
 ;; libpng support.
 
@@ -53,7 +53,8 @@
 (defgroup osm nil
   "OpenStreetMap viewer."
   :link '(info-link :tag "Info Manual" "(osm)")
-  :link '(url-link :tag "Homepage" "https://github.com/minad/osm")
+  :link '(url-link :tag "Website" "https://github.com/minad/osm")
+  :link '(url-link :tag "Wiki" "https://github.com/minad/osm/wiki")
   :link '(emacs-library-link :tag "Library Source" "osm.el")
   :group 'web
   :prefix "osm-")
@@ -152,6 +153,7 @@ Allowed keys:
   :copyright       Copyright information
   :group           Name of server groups for related servers
   :url             Url with placeholders
+  :ext             File name extension
   :min-zoom        Minimum zoom level
   :max-zoom        Maximum zoom level
   :download-batch  Number of tiles downloaded via a single connection
@@ -217,7 +219,9 @@ shown around SVG tiles."
   :type 'symbol)
 
 (defcustom osm-tile-directory
-  (expand-file-name "var/osm/" user-emacs-directory)
+  (expand-file-name (file-name-concat
+                     (or (getenv "XDG_CACHE_HOME") "~/.cache/")
+                     "emacs/osm/"))
   "Tile cache directory."
   :type 'string)
 
@@ -242,6 +246,16 @@ Should be at least 7 days according to the server usage policies."
            (funcall menu)
          menu))))
 
+(defun osm--mouse-ignore-wheel (_prompt)
+  "Ignore mouse wheel events during key translation."
+  (pcase (this-single-command-raw-keys)
+    ((and `[,e]
+          (let y (event-basic-type e))
+          (guard (symbolp y))
+          (guard (string-search "wheel-" (symbol-name y))))
+     [])
+    (k k)))
+
 (defvar-keymap osm-prefix-map
   :doc "Global prefix map of OSM entry points."
   "h" #'osm-home
@@ -258,19 +272,6 @@ Should be at least 7 days according to the server usage policies."
 (defvar-keymap osm-mode-map
   :doc "Keymap used by `osm-mode'."
   :parent (make-composed-keymap osm-prefix-map special-mode-map)
-  "<osm-selected>" #'osm-mouse-select
-  "<osm-bookmark> <mouse-1>" #'osm-mouse-select
-  "<osm-bookmark> <mouse-2>" #'osm-mouse-select
-  "<osm-bookmark> <mouse-3>" #'osm-mouse-select
-  "<osm-home> <mouse-1>" #'osm-mouse-select
-  "<osm-home> <mouse-2>" #'osm-mouse-select
-  "<osm-home> <mouse-3>" #'osm-mouse-select
-  "<osm-poi> <mouse-1>" #'osm-mouse-select
-  "<osm-poi> <mouse-2>" #'osm-mouse-select
-  "<osm-poi> <mouse-3>" #'osm-mouse-select
-  "<osm-track> <mouse-1>" #'osm-mouse-select
-  "<osm-track> <mouse-2>" #'osm-mouse-select
-  "<osm-track> <mouse-3>" #'osm-mouse-select
   "<home>" #'osm-home
   "+" #'osm-zoom-in
   "-" #'osm-zoom-out
@@ -318,6 +319,11 @@ Should be at least 7 days according to the server usage policies."
   "<" nil
   ">" nil)
 
+(dolist (pin osm-pin-colors)
+  (setq pin (vector (car pin)))
+  (define-key key-translation-map pin #'osm--mouse-ignore-wheel)
+  (define-key osm-mode-map pin #'osm-mouse-select))
+
 (easy-menu-define osm-mode-menu osm-mode-map
   "Menu for `osm-mode'."
   '("OSM"
@@ -538,9 +544,10 @@ Local per buffer since the overlays depend on the zoom level.")
    (symbol-name osm-server)
    (format "%d-%d-%d.%s"
            zoom x y
-           (file-name-extension
-            (url-file-nondirectory
-             (osm--server-property :url))))))
+           (or (osm--server-property :ext)
+               (file-name-extension
+                (url-file-nondirectory
+                 (osm--server-property :url)))))))
 
 (defun osm--enqueue-download (x y)
   "Enqueue tile X/Y for download."
@@ -631,6 +638,7 @@ Local per buffer since the overlays depend on the zoom level.")
 
 (defun osm-mouse-drag (event)
   "Handle drag EVENT."
+  (declare (completion ignore))
   (interactive "@e")
   (pcase-let* ((`(,sx . ,sy) (posn-x-y (event-start event)))
                (win (selected-window))
@@ -665,7 +673,7 @@ Local per buffer since the overlays depend on the zoom level.")
 
 (defun osm-center ()
   "Center to location of selected pin."
-  (interactive)
+  (interactive nil osm-mode)
   (osm--barf-unless-osm)
   (pcase osm--pin
     (`(,lat ,lon ,_id ,name)
@@ -684,6 +692,7 @@ Local per buffer since the overlays depend on the zoom level.")
 
 (defun osm-mouse-track (event)
   "Set track pin at location of the click EVENT."
+  (declare (completion ignore))
   (interactive "@e")
   (pcase osm--pin
     ((and (guard (not osm--track)) `(,lat ,lon ,_id ,_name))
@@ -735,22 +744,25 @@ Local per buffer since the overlays depend on the zoom level.")
 
 (defun osm-mouse-pin (event)
   "Create location pin at the click EVENT."
+  (declare (completion ignore))
   (interactive "@e")
   (osm--set-pin-event event)
   (osm--update))
 
 (defun osm-mouse-select (event)
   "Select pin at position of click EVENT."
+  (declare (completion ignore))
   (interactive "@e")
-  (pcase (osm--pin-at event)
-    (`(,lat ,lon ,id ,name)
-     (osm--set-pin id lat lon name (eq id 'osm-track))
-     (when (eq id 'osm-track) (osm--track-length))
-     (osm--update))))
+  (when (memq (event-basic-type event) '(mouse-1 mouse-2 mouse-3))
+    (pcase (osm--pin-at event)
+      (`(,lat ,lon ,id ,name)
+       (osm--set-pin id lat lon name (eq id 'osm-track))
+       (when (eq id 'osm-track) (osm--track-length))
+       (osm--update)))))
 
 (defun osm-zoom-in (&optional n)
   "Zoom N times into the map."
-  (interactive "p")
+  (interactive "p" osm-mode)
   (osm--barf-unless-osm)
   (setq osm--zoom (max (osm--server-property :min-zoom)
                        (min (osm--server-property :max-zoom)
@@ -759,7 +771,7 @@ Local per buffer since the overlays depend on the zoom level.")
 
 (defun osm-zoom-out (&optional n)
   "Zoom N times out of the map."
-  (interactive "p")
+  (interactive "p" osm-mode)
   (osm-zoom-in (- (or n 1))))
 
 (defun osm--move (dx dy)
@@ -769,47 +781,47 @@ Local per buffer since the overlays depend on the zoom level.")
         osm--lat (osm--y-to-lat (+ (osm--y) dy) osm--zoom)))
 
 (defun osm-right (&optional n)
-  "Move N small stepz to the right."
-  (interactive "p")
+  "Move N small steps to the right."
+  (interactive "p" osm-mode)
   (osm--move (* (or n 1) osm-small-step) 0)
   (osm--update))
 
 (defun osm-down (&optional n)
-  "Move N small stepz down."
-  (interactive "p")
+  "Move N small steps down."
+  (interactive "p" osm-mode)
   (osm--move 0 (* (or n 1) osm-small-step))
   (osm--update))
 
 (defun osm-up (&optional n)
-  "Move N small stepz up."
-  (interactive "p")
+  "Move N small steps up."
+  (interactive "p" osm-mode)
   (osm-down (- (or n 1))))
 
 (defun osm-left (&optional n)
-  "Move N small stepz to the left."
-  (interactive "p")
+  "Move N small steps to the left."
+  (interactive "p" osm-mode)
   (osm-right (- (or n 1))))
 
 (defun osm-right-right (&optional n)
-  "Move N large stepz to the right."
-  (interactive "p")
+  "Move N large steps to the right."
+  (interactive "p" osm-mode)
   (osm--move (* (or n 1) osm-large-step) 0)
   (osm--update))
 
 (defun osm-down-down (&optional n)
-  "Move N large stepz down."
-  (interactive "p")
+  "Move N large steps down."
+  (interactive "p" osm-mode)
   (osm--move 0 (* (or n 1) osm-large-step))
   (osm--update))
 
 (defun osm-up-up (&optional n)
-  "Move N large stepz up."
-  (interactive "p")
+  "Move N large steps up."
+  (interactive "p" osm-mode)
   (osm-down-down (- (or n 1))))
 
 (defun osm-left-left (&optional n)
-  "Move N large stepz to the left."
-  (interactive "p")
+  "Move N large steps to the left."
+  (interactive "p" osm-mode)
   (osm-right-right (- (or n 1))))
 
 (defun osm--purge-directory ()
@@ -841,8 +853,7 @@ Local per buffer since the overlays depend on the zoom level.")
         (push (format "%s support" type) req)))
     (unless (libxml-available-p)
       (push "libxml" req))
-    ;; json-available-p is not available on Emacs 27
-    (unless (ignore-errors (equal [] (json-parse-string "[]")))
+    (unless (json-available-p)
       (push "libjansson" req))
     (when req
       (error "Osm: Please compile Emacs with the required libraries, %s needed to proceed"
@@ -868,6 +879,10 @@ Local per buffer since the overlays depend on the zoom level.")
               revert-buffer-function #'osm--revert
               mode-line-process '(:eval (osm--download-queue-info))
               mode-line-position nil
+              mode-line-modified nil
+              mode-line-mule-info nil
+              mode-line-remote nil
+              default-directory (expand-file-name "~/")
               eldoc-documentation-functions nil
               mouse-wheel-progressive-speed nil
               mwheel-scroll-up-function #'osm--zoom-out-wheel
@@ -1035,16 +1050,7 @@ c53 0 96 43 96 96S309 256 256 256z'/>
                   (concat "<svg width='256' height='256' version='1.1'
 xmlns='http://www.w3.org/2000/svg' xmlns:xlink='http://www.w3.org/1999/xlink'>
 <image xlink:href='"
-                          (static-if (> emacs-major-version 27)
-                              (file-name-nondirectory file)
-                            ;; On Emacs 27, :base-uri and embedding by file path
-                            ;; is not supported. Use the less efficient base64
-                            ;; encoding.
-                            (svg--image-data
-                             file
-                             (if (member (file-name-extension file) '("jpg" "jpeg"))
-                                 "image/jpeg" "image/png")
-                             nil))
+                          (file-name-nondirectory file)
                           "' height='256' width='256'/>"
                           (when-let (track (cdr overlays))
                             (format
@@ -1475,7 +1481,7 @@ When called interactively, call the function `osm-home'."
 
 (defun osm-bookmark-set ()
   "Create osm bookmark."
-  (interactive)
+  (interactive nil osm-mode)
   (osm--barf-unless-osm)
   (unwind-protect
       (pcase-let* ((`(,lat ,lon ,loc) (osm--fetch-location-data "New Bookmark"))
@@ -1544,7 +1550,7 @@ When called interactively, call the function `osm-home'."
 
 (defun osm-delete ()
   "Delete selected pin (bookmark or way point)."
-  (interactive)
+  (interactive nil osm-mode)
   (osm--barf-unless-osm)
   (pcase (caddr osm--pin)
     ('nil nil)
@@ -1554,7 +1560,7 @@ When called interactively, call the function `osm-home'."
 
 (defun osm-rename ()
   "Rename selected pin (bookmark or way point)."
-  (interactive)
+  (interactive nil osm-mode)
   (osm--barf-unless-osm)
   (pcase (caddr osm--pin)
     ('osm-bookmark (osm-bookmark-rename (cadddr osm--pin)))
@@ -1571,15 +1577,12 @@ When called interactively, call the function `osm-home'."
                            pins)))
     (pcase (assoc (completing-read
                    "Jump: "
-                     (lambda (str pred action)
-                       (if (eq action 'metadata)
-                           `(metadata
-                             (group-function
-                              . ,(lambda (pin transform)
-                                   (if transform pin
-                                     (cadr (assoc pin pins))))))
-                         (complete-with-action action pins str pred)))
-                     nil t nil 'osm--jump-history)
+                   (osm--table-with-metadata
+                    pins `((group-function
+                            . ,(lambda (pin transform)
+                                 (if transform pin
+                                   (cadr (assoc pin pins)))))))
+                   nil t nil 'osm--jump-history)
                   pins)
       (`(,name ,_group ,id ,lat ,lon ,zoom) (osm--goto lat lon zoom nil id name))
       (_ (user-error "No pin selected")))))
@@ -1622,8 +1625,7 @@ See `osm-search-server' and `osm-search-language' for customization."
    (list
     (minibuffer-with-setup-hook
         (lambda ()
-          (when (and (eq completing-read-function #'completing-read-default)
-                     (not (bound-and-true-p vertico-mode)))
+          (when (eq (keymap-local-lookup "SPC") #'minibuffer-complete-word)
             ;; Override dreaded `minibuffer-complete-word' for default
             ;; completion.  When will this keybinding finally get removed from
             ;; default completion?
@@ -1631,7 +1633,9 @@ See `osm-search-server' and `osm-search-language' for customization."
                             (define-keymap "SPC" nil)
                             (current-local-map)))))
       (completing-read "Location: "
-                       (osm--sorted-table osm--search-history)
+                       (osm--table-with-metadata
+                        osm--search-history '((display-sort-function . identity)
+                                              (cycle-sort-function . identity)))
                        nil nil nil 'osm--search-history))
     current-prefix-arg))
   ;; TODO: Add search bounded to current viewbox, bounded=1, viewbox=x1,y1,x2,y2
@@ -1651,7 +1655,9 @@ See `osm-search-server' and `osm-search-language' for customization."
                       (minibuffer-completion-help))))
               (completing-read
                (format "Matches for '%s': " needle)
-               (osm--sorted-table results)
+               (osm--table-with-metadata
+                results '((display-sort-function . identity)
+                          (cycle-sort-function . identity)))
                nil t nil t))
             results)
            (error "No selection"))))
@@ -1659,13 +1665,13 @@ See `osm-search-server' and `osm-search-language' for customization."
                (apply #'osm--boundingbox-to-zoom (cdddr selected))
                nil 'osm-selected (car selected))))
 
-(defun osm--sorted-table (coll)
-  "Sorted completion table from COLL."
-  (lambda (str pred action)
+;; TODO: Use `completion-table-with-metadata'
+(defun osm--table-with-metadata (table metadata)
+  "Return new completion TABLE with METADATA."
+  (lambda (string pred action)
     (if (eq action 'metadata)
-        '(metadata (display-sort-function . identity)
-                   (cycle-sort-function . identity))
-      (complete-with-action action coll str pred))))
+        `(metadata . ,metadata)
+      (complete-with-action action table string pred))))
 
 ;;;###autoload
 (defun osm-gpx-show (file)
@@ -1676,6 +1682,10 @@ See `osm-search-server' and `osm-search-language' for customization."
                (insert-file-contents file)
                (libxml-parse-xml-region (point-min) (point-max))))
         (min-lat 90) (max-lat -90) (min-lon 180) (max-lon -180))
+    (unless (eq 'gpx (dom-tag dom))
+      (setq dom (dom-child-by-tag dom 'gpx)))
+    (unless (and dom (eq 'gpx (dom-tag dom)))
+      (error "Not a GPX file"))
     (setf (alist-get (abbreviate-file-name file) osm--gpx-files nil nil #'equal)
           (cons
            (cl-loop
@@ -1757,13 +1767,11 @@ See `osm-search-server' and `osm-search-language' for customization."
             osm-server-list))
           (selected (completing-read
                      "Server: "
-                     (lambda (str pred action)
-                       (if (eq action 'metadata)
-                           `(metadata
-                             (annotation-function
-                              . ,(and osm-copyright #'osm--server-annotation))
-                             (group-function . ,#'osm--server-group))
-                         (complete-with-action action servers str pred)))
+                     (osm--table-with-metadata
+                      servers
+                      `((annotation-function
+                         . ,(and osm-copyright #'osm--server-annotation))
+                        (group-function . ,#'osm--server-group)))
                      nil t nil 'osm--server-history
                      (format fmt
                              (osm--server-property :name)
@@ -1777,7 +1785,7 @@ See `osm-search-server' and `osm-search-language' for customization."
 (defun osm-save-url (&optional arg)
   "Save coordinates as url in the kill ring.
 If prefix ARG is given, store url as Elisp expression."
-  (interactive "P")
+  (interactive "P" osm-mode)
   (osm--barf-unless-osm)
   (pcase-let* ((`(,lat ,lon ,loc) (osm--fetch-location-data "New Link"))
                (server (and (not (eq osm-server (default-value 'osm-server))) osm-server))
@@ -1795,13 +1803,13 @@ If prefix ARG is given, store url as Elisp expression."
 
 (cl-defun osm-add-server (server
                           &rest properties
-                          &key name description group url max-connections
+                          &key name description group url ext max-connections
                           max-zoom min-zoom download-batch subdomains copyright)
   "Add SERVER with PROPERTIES to `osm-server-list'.
 The properties are checked as keyword arguments.  See
 `osm-server-list' for documentation of the keywords."
   (declare (indent 1))
-  (ignore name description group url max-connections max-zoom
+  (ignore name description group url ext max-connections max-zoom
           min-zoom download-batch subdomains copyright)
   (dolist (sym '(:name :description :group :url))
     (unless (stringp (plist-get properties sym))
@@ -1812,8 +1820,7 @@ The properties are checked as keyword arguments.  See
   nil)
 
 ;;;###autoload
-(when (>= emacs-major-version 28)
-  (add-to-list 'browse-url-default-handlers '("\\`geo:" . osm)))
+(add-to-list 'browse-url-default-handlers '("\\`geo:" . osm))
 
 ;;;###autoload
 (eval-after-load 'ol
@@ -1827,13 +1834,5 @@ The properties are checked as keyword arguments.  See
               (when (eq major-mode 'osm-mode)
                 (apply 'org-link-store-props (osm--org-link-props)))))))
 
-(dolist (sym (list #'osm-center #'osm-up #'osm-down #'osm-left #'osm-right
-                   #'osm-up-up #'osm-down-down #'osm-left-left #'osm-right-right
-                   #'osm-zoom-out #'osm-zoom-in #'osm-bookmark-set
-                   #'osm-save-url #'osm-rename #'osm-delete))
-  (put sym 'command-modes '(osm-mode)))
-(dolist (sym (list #'osm-mouse-drag #'osm-mouse-pin #'osm-mouse-select #'osm-mouse-track))
-  (put sym 'completion-predicate #'ignore))
-
 (provide 'osm)
 ;;; osm.el ends here
